/*
 * go-font-combo.c :
 *
 * Copyright (C) 2005 Robert Staudinger <robsta@stereolyzer.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/*
 * A font selector combo widget.  
 */

#include <goffice/goffice-config.h>
#include <goffice/gtk/go-combo-font.h>
#include <goffice/gtk/goffice-gtk.h>
#include <goffice/utils/go-font.h>

#include <gtk/gtkcombobox.h>
#include <gtk/gtkcelllayout.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcellrendererpixbuf.h>

#include <gsf/gsf-impl-utils.h>
#include <gsf/gsf-utils.h>
#include <gsf/gsf-input-stdio.h>
#include <gsf/gsf-infile.h>
#include <gsf/gsf-infile-zip.h>
#include <gsf/gsf-output-stdio.h>
#include <gsf/gsf-outfile.h>
#include <gsf/gsf-outfile-zip.h>

#include <gdk/gdkpixbuf.h>
#include <pango/pangoft2.h>

#include <glib/gi18n.h>
#include <glib/gstdio.h>

#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>


#define GOFFICE_CONFIG_DIR ".goffice"
#define CACHEFILE "font-thumbs.cache"
#define CACHEFILE_METASTREAM "__go-combo-font-cache-meta-stream__"
#define KEYFILE_GROUP_DEFAULT "default"
#define KEYFILE_KEY_FONTSIZE "font-size"


struct _GOComboFont {
	GtkComboBox	combo;

	GtkTreeModel *model;
	GtkTreeIter iter;
	PangoFontMap *font_map;
	PangoContext *context;
	gint fontsize;
	
	gboolean cache_dirty;
};

typedef struct {
	GtkComboBoxClass parent_class;

} GOComboFontClass;

enum {
	COLUMN_FAMILY = 0,
	COLUMN_PREVIEW,
	COLUMN_FAMILY_VISIBLE,
	COLUMN_PREVIEW_VISIBLE,
	COLUMN_XALIGN,
	NUM_COLUMNS
};

enum {
	PROP_0, 
	PROP_FONT_FAMILY, 
	PROP_FONT_DESCRIPTION
};

static GtkObjectClass *gcf_parent_class;


/*
* Adapted from
* http://cvs.gnome.org/viewcvs/luminocity/src/decoration.c?rev=1.2&view=markup
*/
static void
get_layout_bitmap (PangoLayout    *layout,
				   FT_Bitmap      *bitmap,
				   PangoRectangle *log_pixels)
{
	PangoRectangle log_rect;

	pango_layout_get_extents (layout, NULL, &log_rect);

	log_pixels->x = log_rect.x >> 10;
	log_pixels->width = ((log_rect.x + log_rect.width + 1023) >> 10) - log_pixels->x;

	log_pixels->y = log_rect.y >> 10;
	log_pixels->height = ((log_rect.y + log_rect.height + 1023) >> 10) - log_pixels->y;

	if (log_pixels->width == 0)
		log_pixels->width = 1;
	if (log_pixels->height == 0)
		log_pixels->height = 1;

	bitmap->width = log_pixels->width;
	bitmap->pitch = (bitmap->width + 3) & ~3;
	bitmap->rows = log_pixels->height;
	bitmap->buffer = g_malloc (bitmap->pitch * bitmap->rows);
	bitmap->num_grays = 256;
	bitmap->pixel_mode = FT_PIXEL_MODE_GRAY;
	memset (bitmap->buffer, 0x00, bitmap->pitch * bitmap->rows);

	pango_ft2_render_layout (bitmap, layout, - log_pixels->x, - log_pixels->y);
}

/*
* Only renders on pixbufs with alpha channel.
* Adapted from
* http://cvs.gnome.org/viewcvs/luminocity/src/decoration.c?rev=1.2&view=markup
*/
static gboolean
draw_layout_on_pixbuf (PangoLayout *layout,
					   GdkPixbuf   *pixbuf,
					   int          x,
					   int          y,
					   int          clip_x,
					   int          clip_y,
					   int          clip_width,
					   int          clip_height)
{
	guchar *pixels = gdk_pixbuf_get_pixels (pixbuf);
	int rowstride = gdk_pixbuf_get_rowstride (pixbuf);
	int bpp = gdk_pixbuf_get_has_alpha (pixbuf) ? 4 : 3;
	guchar *bits;

	PangoRectangle log_pixels;
	FT_Bitmap bitmap;
	int i, j;

	if (!gdk_pixbuf_get_has_alpha (pixbuf)) {
		g_warning ("cannot render on pixbuf without alpha channel");
		return FALSE;
	}

	get_layout_bitmap (layout, &bitmap, &log_pixels);

	bits = bitmap.buffer;

	for (j = y + log_pixels.y; j < y + log_pixels.y + log_pixels.height; j++) {

		if (j >= clip_y && j < clip_y + clip_height) {

			int start_x, end_x;

			start_x = MAX (x + log_pixels.x, clip_x);
			end_x = MIN (x + log_pixels.x + log_pixels.width, clip_x + clip_width);

			if (start_x < end_x) {

				guchar *b = bits + (start_x - (x + log_pixels.x));
				guchar *p = pixels + j * rowstride + start_x * bpp;

				for (i = start_x ; i < end_x; i++) {

					/*
					* Rendering is only done via the alpha channel. 
					* This should make it easier to re-colour the pixbuf later.
					* Or just set the rgb values in the 3 0s.
					*/		

					p[0] = 0;
					p[1] = 0;
					p[2] = 0;
					/* alpha */
					p[3] = *b;

					p += bpp;
					b++;
				}
			}
		}
		bits += bitmap.pitch;
	}

	g_free (bitmap.buffer);
	return TRUE;
}

static gboolean
write_cache_metadata (GOComboFont  *self, 
					  GsfOutfile   *archive, 
					  GError 	  **err)
{
	GKeyFile 	*keyfile;
	GsfOutput 	*stream;
	gchar 		*buffer;
	gsize 		 size;
	
	g_assert (*err == NULL);
	
	keyfile = g_key_file_new ();
	g_key_file_set_integer (keyfile, KEYFILE_GROUP_DEFAULT, KEYFILE_KEY_FONTSIZE, self->fontsize);
	buffer = NULL;
	buffer = g_key_file_to_data (keyfile, &size, err);
	if (*err) {
		if (buffer) g_free (buffer);
		g_key_file_free (keyfile);
		return FALSE;		
	}
	
	stream = gsf_outfile_new_child (archive, CACHEFILE_METASTREAM, FALSE);
	gsf_output_write (stream, size, buffer);
	gsf_output_close (stream);
	g_object_unref (G_OBJECT (stream));
	stream = NULL;
	g_free (buffer);
	g_key_file_free (keyfile);
	
	return TRUE;
}

static void
write_cache (GOComboFont *self)
{
	GsfOutput  	*output;
	GsfOutfile 	*archive;
	GsfOutput  	*stream;
	gchar 		*cachedir;
	gchar 		*cachefile;
	GError   	*err;
	GtkTreeIter  iter;
	
	cachedir = g_build_filename (g_get_home_dir (), GOFFICE_CONFIG_DIR, NULL);
	if (!g_file_test (cachedir, G_FILE_TEST_IS_DIR)) {
		g_mkdir (cachedir, 0755);	
	}
	
	cachefile = g_build_filename (cachedir, CACHEFILE, NULL);
	if (g_file_test (cachefile, G_FILE_TEST_IS_REGULAR)) {
		g_unlink (cachefile);
	}
	g_free (cachedir);
	cachedir = NULL;
	
	output = NULL;
	archive = NULL;
	stream = NULL;
	err = NULL;
	gsf_init ();
	output = gsf_output_stdio_new (cachefile, &err);
	if (err) goto Cleanup;
	
	archive = GSF_OUTFILE (g_object_new (GSF_OUTFILE_ZIP_TYPE, 
						  "sink", output, 
						  "compression-level", 0, 
						  NULL));
	g_object_unref (G_OBJECT (output));

	write_cache_metadata (self, archive, &err);
	if (err) goto Cleanup;
	
	if (gtk_tree_model_get_iter_first (self->model, &iter)) {

		GdkPixbuf 	*pixbuf;
		gchar 		*family;
		gsize 		 size;
		gchar 		*buffer;
		
		do {
			pixbuf = NULL;
			family = NULL;
			buffer = NULL;
			gtk_tree_model_get (self->model, &iter, 
								COLUMN_FAMILY, &family, 
								COLUMN_PREVIEW, &pixbuf, 
								-1);
			g_assert (family && pixbuf);
			
			stream = gsf_outfile_new_child (archive, family, FALSE);
			gdk_pixbuf_save_to_buffer (pixbuf, &buffer, &size, "png", &err, NULL);
			if (err) break;
			gsf_output_write (stream, size, buffer);
			gsf_output_close (stream);
			g_object_unref (G_OBJECT (stream));
			stream = NULL;
			g_free (buffer);
			g_object_unref (pixbuf);
			
		} while (gtk_tree_model_iter_next (self->model, &iter));	
		
		if (err) {
			if (buffer) g_free (buffer);
			if (pixbuf) g_object_unref (pixbuf);
			if (family) g_free (family);
		}
	}
		
Cleanup: 
	if (err) {
		g_warning ("%s\n", err->message);
		g_error_free (err);
	}
	if (stream) { 
		gsf_output_close (stream);
		g_object_unref (G_OBJECT (stream));
	}
	if (archive) {
		gsf_output_close (GSF_OUTPUT (archive));
		g_object_unref (G_OBJECT (archive));
	}
	if (cachefile) g_free (cachefile);
}

static gboolean
timeout_cb (gpointer data)
{
	GOComboFont 		 *self = GO_COMBO_FONT (data);
	PangoLayout 		 *layout;
	PangoFontDescription *desc;
	GdkPixbuf 			 *pixbuf;
	gchar 				 *family;
	PangoRectangle 		  rect;
	gboolean 			  have_next;
	
	
	family = NULL;
	pixbuf = NULL;
	gtk_tree_model_get (self->model, &(self->iter), 
						COLUMN_FAMILY, &family, 
						COLUMN_PREVIEW, &pixbuf, 	
						-1);

	if (pixbuf &&
		(have_next = gtk_tree_model_iter_next (self->model, &(self->iter)))) {
		return TRUE;
	}
	else if (pixbuf && !have_next) {
		if (self->cache_dirty) {
			write_cache (self);	
		}
		return FALSE;
	}		
	
	layout = pango_layout_new (self->context);
	desc = pango_font_description_new ();
	pango_font_description_set_family (desc, family);
	/* TODO heuristics, fc and xft seem to have a different scale */
	pango_font_description_set_size (desc, self->fontsize * 1.5);
	pango_layout_set_font_description (layout, desc);
	pango_layout_set_text (layout, family, -1);

	pango_layout_get_pixel_extents (layout, NULL, &rect);
	pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, rect.width, rect.height);
	draw_layout_on_pixbuf (layout, pixbuf,
							0, - rect.y, 
							0, 0, 
							rect.width, rect.height);
	pango_font_description_free (desc);
	g_object_unref (layout);

	gtk_list_store_set (GTK_LIST_STORE (self->model), &(self->iter), 
						COLUMN_PREVIEW, pixbuf,
						COLUMN_PREVIEW_VISIBLE, TRUE,
						COLUMN_FAMILY_VISIBLE, FALSE, 
						-1);
	g_object_unref (pixbuf);

	if (gtk_tree_model_iter_next (self->model, &(self->iter))) {
		return TRUE;
	}

	if (self->cache_dirty) {
		write_cache (self);	
	}
	return FALSE;
}

static gboolean
read_cache_metadata (GOComboFont *self, 
					 GsfInfile 	 *archive, 
					 GError 	**err)
{
	GsfInput 	*stream;
	GKeyFile 	*keyfile;
	gsf_off_t 	 size;
	gchar const *buffer;
	
	g_assert (*err == NULL);	

	stream = gsf_infile_child_by_name (archive, CACHEFILE_METASTREAM);
	if (stream == NULL) {
		return TRUE;	
	}
	size = gsf_input_size (stream);
	buffer = NULL;
	buffer = gsf_input_read (stream, size, NULL);
	keyfile = g_key_file_new ();
	if (g_key_file_load_from_data (keyfile, buffer, size, G_KEY_FILE_NONE, err)) {
		
		gint fontsize;
		
		fontsize = g_key_file_get_integer (keyfile, KEYFILE_GROUP_DEFAULT, KEYFILE_KEY_FONTSIZE, err);
		if ((fontsize != self->fontsize) ||
			(*err != NULL)) {
			return TRUE;	
		}		
	}
	else {
		return TRUE;	
	}
	
	g_key_file_free (keyfile);
	g_object_unref (G_OBJECT (stream));
	return FALSE;
}

static GdkPixbuf *
load_cached_pixbuf (GOComboFont  *self,
					GsfInfile 	 *archive, 
					const gchar  *family, 
					GError 		**err)
{
	GsfInput 		*stream;
	gsf_off_t 		 size;
	guint8 const 	*buffer;
	GdkPixbufLoader *loader;
	GdkPixbuf 		*pixbuf;
	
	g_assert (*err == NULL);
	
	if (archive == NULL) {
		return NULL;	
	}
	
	stream = gsf_infile_child_by_name (archive, family);
	if (stream == NULL) {
		/* new fonts on system, recreate cache */
		self->cache_dirty = TRUE;
		return NULL;	
	}
	
	size = gsf_input_size (stream);	
	buffer = gsf_input_read (stream, size, NULL);

	loader = gdk_pixbuf_loader_new_with_type ("png", err);
	gdk_pixbuf_loader_write (loader, buffer, size, err);
	if (*err) {
		g_object_unref (G_OBJECT (stream));
		g_object_unref (G_OBJECT (loader));
		return NULL;	
	}
	
	gdk_pixbuf_loader_close (loader, err);
	if (*err) {
		g_object_unref (G_OBJECT (stream));
		g_object_unref (G_OBJECT (loader));
		return NULL;	
	}

	pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
	g_object_ref (G_OBJECT (pixbuf));
	g_object_unref (G_OBJECT (loader));

	return pixbuf;
}

static void
realize_cb (GtkWidget *widget,
			gpointer   data)
{
	GOComboFont *self = GO_COMBO_FONT (widget);
	PangoContext 		 *context;
	GSList 				 *family_names;
	GSList 				 *ptr;
	PangoFontDescription *desc;
	gchar 				 *cachefile;
	GsfInput 			 *input;
	GsfInfile 			 *archive;
	GsfInput 			 *stream;
	GdkPixbuf 			 *pixbuf;
	GError 				 *err;
	GtkTreeIter 		  iter;
	
	/* font size */
	context = NULL;
	context = gtk_widget_get_pango_context (GTK_WIDGET (self));
	desc = pango_context_get_font_description (context);
	self->fontsize = pango_font_description_get_size (desc);

	family_names = NULL;
	self->font_map = pango_ft2_font_map_new ();
	self->context = pango_ft2_font_map_create_context (PANGO_FT2_FONT_MAP (self->font_map));
	family_names = go_fonts_list_families (self->context);
	
	/* cache file */
	cachefile = NULL;
	cachefile = g_build_filename (g_get_home_dir (), GOFFICE_CONFIG_DIR, CACHEFILE, NULL);
	if (!g_file_test (cachefile, G_FILE_TEST_IS_REGULAR)) {
		self->cache_dirty = TRUE;
	}	
	
	err = NULL;
	input = NULL;
	archive = NULL;
	stream = NULL;
	gsf_init ();
	input = gsf_input_stdio_new (cachefile, &err);
	if (err) {
		self->cache_dirty = TRUE;
		g_error_free (err);
		err = NULL;
		input = NULL;
	}
	else {
		archive = gsf_infile_zip_new (input, &err);
		if (err) {
			self->cache_dirty = TRUE;
			g_warning ("%s\n", err->message);
			g_error_free (err);
			err = NULL;			
			archive = NULL;
		}
		g_object_unref (G_OBJECT (input));	
	}
	g_free (cachefile);
	cachefile = NULL;

	if (archive) {
		self->cache_dirty = read_cache_metadata (self, archive, &err);
	}	
	
	ptr = NULL;
	for (ptr = family_names; ptr != NULL; ptr = ptr->next) {
		
		if (!self->cache_dirty) {
			pixbuf = load_cached_pixbuf (self, archive, (const gchar*)(ptr->data), &err);
		}
		else {
			pixbuf = NULL;
		}

		gtk_list_store_append (GTK_LIST_STORE (self->model), &iter);
		gtk_list_store_set (GTK_LIST_STORE (self->model), &iter, 
							COLUMN_FAMILY, ptr->data, 
							COLUMN_FAMILY_VISIBLE, !pixbuf, 
							COLUMN_PREVIEW, pixbuf, 
							COLUMN_PREVIEW_VISIBLE, TRUE, 
							COLUMN_XALIGN, 0.0, 
							-1);
		if (pixbuf) {
			g_object_unref (G_OBJECT (pixbuf));
		}
	}

	g_slist_foreach (family_names, (GFunc)g_free, NULL); 
	g_slist_free (family_names); 
	family_names = NULL;

	if (gtk_tree_model_get_iter_first (self->model, &(self->iter)))
		g_timeout_add (1000, timeout_cb, self);
}

static void
gcf_init (GOComboFont *self)
{
	GtkCellRenderer *renderer;
	
	self->cache_dirty = FALSE;
	
	self->model = GTK_TREE_MODEL (gtk_list_store_new (NUM_COLUMNS, 
													  G_TYPE_STRING, GDK_TYPE_PIXBUF,
													  G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, 
													  G_TYPE_FLOAT));
	gtk_combo_box_set_model (GTK_COMBO_BOX (self), self->model);
	g_object_unref (self->model);
	
	renderer = gtk_cell_renderer_text_new ();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (self), renderer, TRUE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (self), renderer, "text", COLUMN_FAMILY);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (self), renderer, "visible", COLUMN_FAMILY_VISIBLE);

	renderer = gtk_cell_renderer_pixbuf_new ();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (self), renderer, TRUE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (self), renderer, "pixbuf", COLUMN_PREVIEW);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (self), renderer, "visible", COLUMN_PREVIEW_VISIBLE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (self), renderer, "xalign", COLUMN_XALIGN);

	g_signal_connect_after (G_OBJECT (self), "realize", G_CALLBACK (realize_cb), NULL);
}

static void
gcf_destroy (GtkObject *object)
{
	GOComboFont *self = GO_COMBO_FONT (object);

	/* g_object_unref (self->context); TODO this bails */
	g_object_unref (self->font_map);

	gcf_parent_class->destroy (object);
}

static void
get_property (GObject      *object,
			  guint         property_id,
			  GValue       *value,
			  GParamSpec   *pspec)
{
	GOComboFont 		 *self = GO_COMBO_FONT (object);
	gchar 				 *family = NULL;
	PangoFontDescription *desc = NULL;
	GtkTreeIter 		  iter;
	
	if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (self), &iter)) {
		
		gtk_tree_model_get (self->model, &iter, 
							COLUMN_FAMILY, &family, 
							-1);
	}
	
	switch (property_id) {
		case PROP_FONT_FAMILY: 
			g_value_set_string (value, family);
			break;
		case PROP_FONT_DESCRIPTION:
			desc = pango_font_description_new ();
			pango_font_description_set_family (desc, family);
			g_value_set_pointer (value, desc);
		break;
		default: 
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);		
	}
}

static void
set_property (GObject      *object,
			  guint         property_id,
			  const GValue *value,
			  GParamSpec   *pspec)
{
	GOComboFont *self = GO_COMBO_FONT (object);
	const gchar *family = NULL;
	GtkTreeIter  iter;
	
	switch (property_id) {
		case PROP_FONT_FAMILY: 
			family = g_value_get_string (value);
			break;
		case PROP_FONT_DESCRIPTION:
			family = pango_font_description_get_family ((PangoFontDescription*)g_value_get_pointer (value));
			break;
		default: 
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);		
	}

	if (family &&
		gtk_tree_model_get_iter_first (self->model, &iter)) {
		gchar *tmp;
		gint idx = 0;
		do {
			tmp = NULL;
			gtk_tree_model_get (self->model, &iter, 
								COLUMN_FAMILY, &tmp, 
								-1);
			if (0 == strncmp (family, tmp, g_utf8_strlen (family, -1))) {
				gtk_combo_box_set_active (GTK_COMBO_BOX (self), idx);
				g_free (tmp);
				break;
			}
			g_free (tmp);
			idx++;
			
		} while (gtk_tree_model_iter_next (self->model, &iter));
	}
}

static void
gcf_class_init (GObjectClass *klass)
{
	GObjectClass 	*gobject_class = (GObjectClass *) klass;
	GtkObjectClass 	*gto_class = (GtkObjectClass *) klass;

	gto_class->destroy = gcf_destroy;

	gcf_parent_class = g_type_class_peek (gtk_combo_box_get_type ());

	gobject_class->set_property = set_property;
 	gobject_class->get_property = get_property;
	
	g_object_class_install_property
		(gobject_class,
		PROP_FONT_FAMILY,
		g_param_spec_string 
			("font-family",
			_("Font family"),
			_("Name of font family"),
			NULL,
			G_PARAM_READWRITE));

	g_object_class_install_property
		(gobject_class,
		PROP_FONT_DESCRIPTION,
		g_param_spec_pointer 
			("font-description",
			_("Font description"),
			_("Pango font description of font"),
			G_PARAM_READWRITE));
}

GSF_CLASS (GOComboFont, go_combo_font,
	   gcf_class_init, gcf_init, GTK_TYPE_COMBO_BOX)

GtkWidget *
go_combo_font_new (void)
{
	return g_object_new (GO_COMBO_FONT_TYPE, NULL);
}

void	      
go_combo_font_set_description (GOComboFont 			*self, 
							   PangoFontDescription *desc)
{
	GValue value = {0,};
	
	g_value_init (&value, G_TYPE_POINTER);
	g_value_set_pointer (&value, desc);
	g_object_set_property (G_OBJECT (self), "font-description", &value);
}

PangoFontDescription const *
go_combo_font_get_description (GOComboFont const *self)
{
	GValue value = {0,};
	
	g_value_init (&value, G_TYPE_POINTER);
	g_object_get_property (G_OBJECT (self), "font-description", &value);	
	return (PangoFontDescription*)g_value_get_pointer (&value);
}

void	      
go_combo_font_set_family (GOComboFont *self, 
						  gchar 	  *family)
{
	GValue value = {0,};
	
	g_value_init (&value, G_TYPE_STRING);
	g_value_set_string (&value, family);
	g_object_set_property (G_OBJECT (self), "font-family", &value);	
}

gchar *		  
go_combo_font_get_family (GOComboFont const *self)
{
	GValue value = {0,};
	
	g_value_init (&value, G_TYPE_STRING);
	g_object_get_property (G_OBJECT (self), "font-family", &value);	
	return (gchar *)g_value_get_string (&value);	
}
